from asyncio.subprocess import Process
import os
from .load_exe import get_unzipped_host_path
from ..utils.run_shell_cmd import run_shell_cmd
from ..utils.globals import globals
import signal
import webbrowser
from ..utils.wait_till_file_server_ready import wait_till_file_server_ready

async def open_exe(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir) -> Process:
    abs_app_dir = os.path.abspath(app_dir)
    child_proc = await run_shell_cmd(str(get_unzipped_host_path()), os.getcwd(), not globals["in_debug_mode"])
    return child_proc

async def open_exe_wrapper(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir):
    child_proc = await open_exe(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir)
    # terminate docker process if SIGINT, SIGTERM is received
    def handle_signal(a, b):
        child_proc.terminate()
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    await wait_till_file_server_ready(w_port)
    print("Success! Visit http://localhost:{} to access the editor.".format(w_port))
    webbrowser.open("http://localhost:4002", new=0, autoraise=True)
    await child_proc.wait()



from ..find_app_root import get_virtualenv_type
from ..errors import UNKNOWN_VIRT_TYPE
from .. import supported_virt_types
from .run_shell_cmd import run_shell_cmd
from .globals import globals
from .conda_utils import get_working_env_name

def get_common_command():
    return "python -m controllers.server serve"

async def pipenv_call_serve(app_dir: str):
    cmd = "pipenv run" + " " + get_common_command()
    python_server_proc = await run_shell_cmd(
        cmd,
        app_dir,
        not globals["in_debug_mode"]
        )
    return python_server_proc

async def conda_call_serve(app_dir: str):
    cmd = "conda run -n " + get_working_env_name() + " " + get_common_command()
    python_server_proc = await run_shell_cmd(
        cmd,
        app_dir,
        not globals["in_debug_mode"]
        )
    return python_server_proc

async def call_serve(app_dir: str):
    virt_type = get_virtualenv_type()
    if virt_type == "pipenv":
        return await pipenv_call_serve(app_dir)
    elif virt_type == "conda":
        return await conda_call_serve(app_dir)
    else:
        print("Error:", "Unsupported virtual environment", virt_type)
        print("Please use a virtual environment from this list:")
        print(supported_virt_types)
        exit(UNKNOWN_VIRT_TYPE)


import subprocess
from .run_shell_cmd import run_shell_cmd
import os
from .globals import globals
from ..find_app_root import get_virtualenv_type
from ..errors import UNKNOWN_VIRT_TYPE
from .. import supported_virt_types

async def is_pipenv_installed():
    child_proc = await run_shell_cmd("pipenv --version", os.getcwd(), globals["in_debug_mode"])
    await child_proc.wait()
    if child_proc.returncode == 0:
        return True
    return False

async def is_conda_installed():
    child_proc = await run_shell_cmd("conda --version", os.getcwd(), globals["in_debug_mode"])
    await child_proc.wait()
    if child_proc.returncode == 0:
        return True
    return False

def is_conda_installed_sync() -> bool:
    try:
        child_proc = subprocess.Popen(["conda", "--version"], stderr=subprocess.PIPE, stdout=subprocess.PIPE, stdin=subprocess.DEVNULL)
        child_proc.wait()
        if child_proc.returncode == 0:
            return True
        return False
    except:
        return False

async def is_selected_virtenv_installed():
    virt_type = get_virtualenv_type()
    if virt_type == "pipenv":
        return await is_pipenv_installed()
    elif virt_type == "conda":
        return await is_conda_installed()
    else:
        print("Error:", "Unsupported virtual environment", virt_type)
        print("Please use a virtual environment from this list:")
        print(supported_virt_types)
        exit(UNKNOWN_VIRT_TYPE)


from ..errors import PIP_NOT_INSTALLED, PYTHON_NOT_INSTALLED, SELECTED_VIRTENV_NOT_INSTALLED, DOCKER_NOT_INSTALLED

def error_to_message(error_code: int):
    if error_code == DOCKER_NOT_INSTALLED:
        print("Error: docker not installed. Please install docker.")
    if error_code == SELECTED_VIRTENV_NOT_INSTALLED:
        print("Error: The selected virtual environment is not installed. \
        Please check atri.app.json file to find your selected virtual env type.")
    if error_code == PYTHON_NOT_INSTALLED:
        print("Please install python in your conda environment.")
        print("Run conda install python")
    if error_code == PIP_NOT_INSTALLED:
        print("Please install pip in your conda environment.")
        print("Run conda install pip")   


import requests
import asyncio

async def wait_till_file_server_ready(file_server_port: int):
    while True:
        try:
            requests.get("http://localhost:" + str(file_server_port))
            break
        except:
            print("file server not ready")
            await asyncio.sleep(0.2)


from typing import Union, List
import subprocess
import re

def get_conda_env_list(app_dir: str) -> List[str]:
    """
    Get list of conda env names.
    """
    child_proc = subprocess.Popen(
        ["conda", "env", "list"],
        stdout=subprocess.PIPE,
        cwd=app_dir
        )
    out = child_proc.stdout.read().decode("utf-8")
    # parse lines starting without #
    env_names: List[str] = []
    for line in out.splitlines():
        if not re.search("^#", line):
            for ind, a in enumerate(line):
                if a.isspace():
                    env_name = line[0:ind]
                    env_names.append(env_name)
                    break
    return env_names

def get_active_env_name(app_dir: str) -> Union[str, None]:
    """
    Returns name of the active conda environment if any otherwise None.
    """
    child_proc = subprocess.Popen(
        ["conda", "env", "list"],
        stdout=subprocess.PIPE,
        cwd=app_dir
        )
    out = child_proc.stdout.read().decode("utf-8")
    env_name: Union[str, None] = None
    for line in out.splitlines():
        if not re.search("^#", line):
            temp_env_name: Union[str, None] = None
            for ind, a in enumerate(line):
                # find first space and store the string before it in temp variable
                if a.isspace() and temp_env_name == None:
                    temp_env_name = line[0:ind]
                # find * in the line
                if a == "*" and temp_env_name != None:
                    env_name = temp_env_name
                    break
        if env_name != None:
            break
    return env_name

def is_pkg_installed_in_env(env_name: str, pkg_name: str, app_dir: str) -> bool:
    """
    check if packages like python and pip are installed
    """
    child_proc = subprocess.Popen(
        ["conda", "list", "-n", env_name],
        stdout=subprocess.PIPE,
        cwd=app_dir
        )
    out = child_proc.stdout.read().decode("utf-8")
    # parse lines starting without #
    for line in out.splitlines():
        if not re.search("^#", line):
            for ind, a in enumerate(line):
                if a.isspace():
                    curr_pkg_name = line[0:ind]
                    if curr_pkg_name == pkg_name:
                        return True
                    break
    return False

working_env_name: Union[str, None]
def set_working_env_name(name: str):
    global working_env_name
    working_env_name = name
def get_working_env_name():
    return working_env_name



import asyncio
from ..utils.printd import printd
from ..errors import SELECTED_VIRTENV_NOT_INSTALLED, PYTHON_NOT_INSTALLED, PIP_NOT_INSTALLED
from ..utils.is_pkg_installed import is_selected_virtenv_installed
from ..find_app_root import get_virtualenv_type
from ..utils.conda_utils import is_pkg_installed_in_env, get_working_env_name
from pathlib import Path

async def check_docker_installed():
    child_proc = await asyncio.create_subprocess_shell(
        "docker images",
        stderr=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE
        )
    # call communicate otherwise it might go into deadlock
    # refer https://docs.python.org/3/library/asyncio-subprocess.html
    _, stderr = await child_proc.communicate()
    if child_proc.returncode != 0:
        print(
            "Please install docker.",
            "If docker is already installed, then please make sure that docker daemon is running."
            )
        printd("[stderr]\n", stderr.decode("utf-8"))
        return False
    return True

async def check_selected_virtenv_is_installed():
    return await is_selected_virtenv_installed()

async def check_requisite():
    is_selected_virtenv_installed = await check_selected_virtenv_is_installed()
    if not is_selected_virtenv_installed:
        return SELECTED_VIRTENV_NOT_INSTALLED
    if get_virtualenv_type() == "conda":
        is_python_installed = is_pkg_installed_in_env(get_working_env_name(), "python", str(Path.cwd()))
        if not is_python_installed:
            return PYTHON_NOT_INSTALLED
        is_pip_installed = is_pkg_installed_in_env(get_working_env_name(), "pip", str(Path.cwd()))
        if not is_pip_installed:
            return PIP_NOT_INSTALLED
    return 0


from pathlib import Path
import os

def get_atri_dir_path():
    atri_dir = Path.home() / ".atri"
    return atri_dir

def create_atri_dir_if_not_exist():
    atri_dir = get_atri_dir_path()
    if not os.path.exists(str(atri_dir)):
        os.mkdir(str(atri_dir))


# __version__ is used in docker run command to identify
# the version of docker image to run
__version__ = "0.0.39"

supported_virt_types = ["pipenv", "conda"]

app_config_file = "atri.app.json"

assets_dir = "assets"
atri_templates_dir = "atri_templates"
controllers_dir = "controllers"
atri_app_dir = "atri_app"


from .run_shell_cmd import run_shell_cmd
from ..find_app_root import get_virtualenv_type
from ..errors import UNKNOWN_VIRT_TYPE
from .. import supported_virt_types
from .run_shell_cmd import run_shell_cmd
from .conda_utils import get_working_env_name

async def install_with_pipenv(app_dir: str, pkg: str, version: str):
    full_install_path = pkg if version == "*" else pkg + version
    child_proc = await run_shell_cmd("pipenv install " + full_install_path, app_dir)
    return child_proc

async def install_with_conda(app_dir: str, pkg: str, version: str):
    full_install_path = pkg if version == "*" else pkg + version
    child_proc = await run_shell_cmd("conda run -n " + get_working_env_name() + " python -m pip install " + full_install_path, app_dir)
    return child_proc

async def install_package(app_dir: str, pkg: str, version: str):
    virt_type = get_virtualenv_type()
    if virt_type == "pipenv":
        return await install_with_pipenv(app_dir, pkg, version)
    elif virt_type == "conda":
        return await install_with_conda(app_dir, pkg, version)
    else:
        print("Error:", "Unsupported virtual environment", virt_type)
        print("Please use a virtual environment from this list:")
        print(supported_virt_types)
        exit(UNKNOWN_VIRT_TYPE)


import asyncio
from math import inf
import os
import traceback
import socketio
from ..utils.install_package import install_package
import toml
from ..utils.printd import printd
from ..utils.call_compute import call_compute
from ..utils.call_serve import call_serve
from ..find_app_root import get_virtualenv_type

async def on_connect(sio):
    def print_success_cb(success):
        if success:
            printd("Registered as atri-cli with ipc server.")
    await sio.emit("registerAs", "atri-cli", callback=print_success_cb)

async def connect_ipc_server(port: str):
    printd("Connecting to ipc server...")
    sio = socketio.AsyncClient(
        reconnection=True,
        reconnection_attempts=inf,
        reconnection_delay=1,
        reconnection_delay_max=4,
        handle_sigint=True
        )

    @sio.on('connect')
    async def connect_handler():
        printd('Connected to ipc server!')
        await on_connect(sio)

    @sio.event
    def connect_error(data):
        printd("Connection to ipc server failed!")

    @sio.event
    def disconnect():
        printd("Disconnected from ipc server!")

    while True:
        try:
            await sio.connect("http://localhost:" + port)
            break
        except:
            await asyncio.sleep(1)
    return sio

def handle_ipc_events(sio, paths):
    python_server_proc = None
    @sio.on("doComputeInitialState")
    async def doComputeInitialState(route: str, page_state: str):
        printd("doComputeInitialState called")
        try:
            app_dir = paths["app_dir"]
            out = await call_compute(app_dir, route, page_state)
            return True, out
        except Exception:
            print("except", traceback.print_exc())
    @sio.on("doBuildPython")
    async def doBuildPython():
        printd("doBuildPython called")
        app_dir = paths["app_dir"]
        controllers_dir = os.path.join(app_dir, "controllers")
        initial_pipfile_path = os.path.join(controllers_dir, "Pipfile")
        # check if Pipfile exist in controller directory
        if os.path.exists(initial_pipfile_path):
            # read Pipfile
            pipfile_data = toml.load(initial_pipfile_path)
            pkgs = pipfile_data["packages"]
            dev_pkgs = pipfile_data["dev-packages"]
            # run pipenv install <package_name> for each file in Pipfile inside app_dir
            for pkg in pkgs:
                version = pkgs[pkg]
                if type(version) != str:
                    version = version["version"]
                child_proc = await install_package(app_dir, pkg, version)
                _, stderr = await child_proc.communicate()
                if child_proc.returncode != 0:
                    print("Failed: {} install".format(get_virtualenv_type()), pkg, version)
                    if stderr:
                        printd("[stderr]\n", stderr)
                else:
                    printd("Installed", pkg, version)
            for pkg in dev_pkgs:
                version = dev_pkgs[pkg]
                if type(version) != str:
                    version = version["version"]
                child_proc = await install_package(app_dir, pkg, version)
                _, stderr = await child_proc.communicate()
                if child_proc.returncode != 0:
                    print("Failed: {} install".format(get_virtualenv_type()), pkg, version)
                    if stderr:
                        printd("[stderr]\n", stderr)
                else:
                    printd("Installed", pkg, version)
            # delete Pipfile from controllers_dir
            os.remove(initial_pipfile_path)
        return True
    @sio.on("doStartPythonServer")
    async def doStartPythonServer():
        printd("doStartPythonServer called")
        nonlocal python_server_proc
        app_dir = paths["app_dir"]
        if python_server_proc == None:
            python_server_proc = await call_serve(app_dir)
            await python_server_proc.wait()
            if python_server_proc.returncode != 0:
                print("Error occured while running python -m controllers.server serve")
            python_server_proc = None
async def start_ipc_connection(port: str, app_dir):
    abs_app_dir = os.path.abspath(app_dir)
    paths = {"app_dir": abs_app_dir}
    sio = await connect_ipc_server(port)
    handle_ipc_events(sio, paths)
    return sio

def run(u_port, app_dir):
    async def wrapper():
        # Important to call sio.wait if no other task will run
        # If sio.wait is not called, then the python program will crash in ~30 secs
        # with error message 'packet queue is empty, aborting'
        sio = await start_ipc_connection(u_port, app_dir)
        await sio.wait()
    asyncio.run(wrapper())


from ..find_app_root import get_virtualenv_type
from ..errors import UNKNOWN_VIRT_TYPE
from .. import supported_virt_types
import subprocess
from .conda_utils import get_working_env_name


def get_common_command(route: str, page_state: str):
    common_command = ["python", "-m", "controllers.server", "compute", "--route", route, "--state", page_state]
    return common_command

async def pipenv_call_compute(app_dir: str, route: str, page_state: str):
    child_proc = subprocess.Popen(
        ["pipenv", "run", *get_common_command(route, page_state)],
        stdout=subprocess.PIPE,
        cwd=app_dir
        )
    out = child_proc.stdout.read()
    return out

async def conda_call_compute(app_dir: str, route: str, page_state: str):
    child_proc = subprocess.Popen(
        ["conda", "run", "-n", get_working_env_name(), *get_common_command(route, page_state)],
        stdout=subprocess.PIPE,
        cwd=app_dir
        )
    out = child_proc.stdout.read()
    return out

async def call_compute(app_dir: str, route: str, page_state: str):
    virt_type = get_virtualenv_type()
    if virt_type == "pipenv":
        return await pipenv_call_compute(app_dir, route, page_state)
    elif virt_type == "conda":
        return await conda_call_compute(app_dir, route, page_state)
    else:
        print("Error:", "Unsupported virtual environment", virt_type)
        print("Please use a virtual environment from this list:")
        print(supported_virt_types)
        exit(UNKNOWN_VIRT_TYPE)


# copied from stackoverflow https://stackoverflow.com/questions/1871549/determine-if-python-is-running-inside-virtualenv
import sys

def get_base_prefix_compat():
    """Get base/real prefix, or sys.prefix if there is none."""
    return getattr(sys, "base_prefix", None) or getattr(sys, "real_prefix", None) or sys.prefix

def in_virtualenv():
    return get_base_prefix_compat() != sys.prefix


from pathlib import Path
import os
import json
import sys
from . import __version__, app_config_file, supported_virt_types
from .utils.printd import printd

def find_atri_root_dir():
    curr_dir = Path.cwd()
    while True:
        attempt = curr_dir / app_config_file
        if os.path.exists(str(attempt)):
            return curr_dir
        # already at root, cannot go up anymore
        if str(curr_dir) == str(curr_dir.parent):
            return None
        curr_dir = curr_dir.parent

    return None

def find_and_set_app_directory():
    """
    NOTE: use it only inside cli.py or any other entry module.
    This function sets app's root directory as the cwd for the
    running python process.

    It creates atri.app.json file in cwd if not found.
    """
    dir = find_atri_root_dir()
    if dir == None:
        with open(app_config_file, "w") as f:
            content = {"version": __version__}
            f.write(json.dumps(content, indent=2))
    else:
        os.chdir(dir)

def is_supported_virtualenv_type(virt_type):
    return virt_type in supported_virt_types

def get_virtualenv_type():
    try:
        with open(app_config_file, "r") as f:
            content = json.loads(f.read())
            if "virt_type" in content:
                # check if it's a supported virtual env
                virt_type = content["virt_type"]
                if is_supported_virtualenv_type(virt_type):
                    return virt_type
                else:
                    return None
            return None
    except:
        print("Problem reading", app_config_file, "in", Path.cwd())
        printd("[Error]")
        printd(sys.exc_info())
        return None

def is_virtualenv_set():
    if get_virtualenv_type() != None:
        return True
    else:
        return False

def set_virtualenv_type(virt_type: str):
    """
    This function should only be called after find_and_set_app_directory
    has been called. In other words, cwd must be set correctly before
    calling this function. This function also assumes that atri.app.json
    file exists.
    """
    old_content = ""
    with open(app_config_file, "r+") as f:
        old_content = json.loads(f.read())
        f.truncate(0)
    with open(app_config_file, "r+") as f:
        new_content = {"virt_type": virt_type}
        new_content.update(old_content)
        f.write(json.dumps(new_content, indent=2))



import os
import asyncio
from ..utils.run_shell_cmd import run_shell_cmd
from .. import __version__
from ..utils.globals import globals
import signal
import webbrowser

def port_map(host: str, container: str):
    return host + ":" + container

def volume_map(app_dir:str, subdir: str, container_dir: str):
    return os.path.join(app_dir, subdir) + ":" + container_dir

def create_env(name: str, value: str):
    return name + "=" + value

def host_map(container: str, host: str):
    return container + ":" + host

async def open_editor(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir):
    abs_app_dir = os.path.abspath(app_dir)
    host_uri_inside_docker = "host.docker.internal"
    atri_controller_host = host_uri_inside_docker + ":" + c_port
    cmd = " ".join([
        "docker", "run", "--rm",
        "-p", port_map(e_port, "4001"),
        "-p", port_map(w_port, "4002"),
        "-p", port_map(m_port, "4003"),
        "-p", port_map(p_port, "4004"),
        "-p", port_map(d_port, "4005"),
        "-p", port_map(u_port, "4006"),
        "-e", create_env("ATRI_CONTROLLER_HOST", atri_controller_host),
        "--add-host", host_map(host_uri_inside_docker, "host-gateway"),
        "-v", volume_map(abs_app_dir, "localdb", "/code/localdb"),
        "-v", volume_map(abs_app_dir, "controllers", "/code/node_modules/.targets/controllers"),
        "-v", volume_map(abs_app_dir, "atri-app", "/code/node_modules/.targets/atri-app"),
        "-v", volume_map(abs_app_dir, "assets", "/code/node_modules/.targets/assets"),
        "atrilabs/atrieditor:" + __version__
        ])
    child_proc = await run_shell_cmd(cmd, os.getcwd(), not globals["in_debug_mode"])
    return child_proc

def run(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir):
    try:
        # create a child process
        async def wrapper():
            child_proc = await open_editor(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir)
            await child_proc.wait()
        asyncio.run(wrapper())
    except KeyboardInterrupt:
        return
    except Exception as e:
        if e.args[0] != 10:
            print(e)

async def open_editor_wrapper(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir):
    child_proc = await open_editor(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir)
    # terminate docker process if SIGINT, SIGTERM is received
    def handle_signal(a, b):
        child_proc.terminate()
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    print("Success! Visit http://localhost:4002 to access the editor.")
    await asyncio.sleep(2)
    webbrowser.open("http://localhost:4002", new=0, autoraise=True)
    await child_proc.wait()


import asyncio

async def run_shell_cmd(cmd: str, cwd: str, pipe: bool = True):
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE if pipe else None,
        stderr=asyncio.subprocess.PIPE if pipe else None,
        stdin=asyncio.subprocess.DEVNULL if pipe else None,
        cwd=cwd
        )
    return proc


import sys
import setuptools
import os

EDITABLE_BUILD = os.getenv("PIP_EDITABLE_BUILD", None) == "true"

try:
    from pipenv.project import Project
    try:
        # Pipenv 2022.4.8
        from pipenv.utils.dependencies import convert_deps_to_pip
    except:
        # Older Pipenv
        from pipenv.utils import convert_deps_to_pip
except:
    err_msg = "Please install pipenv and try again."
    sys.exit(err_msg)

import subprocess
import re

def get_git_tag():
    try:
        git_tag = str(
            subprocess.check_output(
                ['git', 'describe', '--exact-match', '--abbrev=0'], stderr=subprocess.STDOUT
            )
        ).strip('\'b\\n')
    except subprocess.CalledProcessError as exc_info:
        match = re.search(
            'fatal: no tag exactly matches \'(?P<commit>[a-z0-9]+)\'', str(exc_info.output)
        )
        if match:
            raise Exception(
                'Bailing: there is no git tag for the current commit, {commit}'.format(
                    commit=match.group('commit')
                )
            )
        raise Exception(str(exc_info.output))

    return git_tag

NAME = "atri"

VERSION = "1.0.0" if EDITABLE_BUILD else get_git_tag()[1:]

DESCRIPTION = "The best tools to build better and fast apps"

LONG_DESCRIPTION = (
    "A full stack framework that let's you drag-&-drop components to build apps."
    "The backend of these apps can be very easily built using Python."
    "It (semi-)automates many things out of the box such as CDN caching, browser caching etc."
)

pipfile = Project().parsed_pipfile
packages = pipfile["packages"].copy()
requirements = convert_deps_to_pip(packages, r=False)

setuptools.setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://atrilabs.com",
    project_urls={
        "Source": "https://github.com/cruxcode/atrilabs-engine.git"
    },
    author="Atri Labs",
    author_email="shyam.swaroop@atrilabs.com",
    python_requires=">=3.7",
    license="GPLv3",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=requirements,
    zip_safe=False,
    include_package_data=True,
    entry_points={"console_scripts": ["atri=atri.cli:main"]}
)


from .. import __version__
import platform
import requests
from ..utils.atri_dir_utils import get_atri_dir_path, create_atri_dir_if_not_exist
import zipfile
from pathlib import Path
import os
import sys

def get_remote_filename():
    if platform.system() == "Linux":
        system_str = "linux"
    if platform.system() == "Darwin":
        system_str = "macos"
    if platform.system() == "Windows":
        system_str = "win.exe"
    filename = "webapp-builder-" + system_str + ".zip"
    return filename

def get_host_filename():
    if platform.system() == "Linux":
        system_str = "linux"
    if platform.system() == "Darwin":
        system_str = "macos"
    if platform.system() == "Windows":
        system_str = "win.exe"
    filename = "webapp-builder-" + system_str + ".v" + __version__ + ".zip"
    return filename

def download_zip():
    atri_dir_path = get_atri_dir_path()
    remote_filename = get_remote_filename()
    host_filename = get_host_filename()
    full_filepath_in_host = atri_dir_path / host_filename

    url = "https://github.com/Atri-Labs/atrilabs-engine/releases/download/" \
        + "v" + __version__ + "/" + remote_filename
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(full_filepath_in_host, "wb") as f:
            count = 1
            chunk_size = 81960
            total_size = int(r.headers["Content-Length"])
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)
                progress = (count * len(chunk) / total_size) * 100
                # erase and go to beginning of line
                # https://stackoverflow.com/questions/5290994/remove-and-replace-printed-items
                sys.stdout.write("\033[K")
                print("Download Progress", str(round(progress, 2)) + "%", end="\r")
                count = count + 1
            sys.stdout.write("\033[K")

def get_unzipped_remote_filename():
    if platform.system() == "Linux":
        system_str = "linux"
    if platform.system() == "Darwin":
        system_str = "macos"
    if platform.system() == "Windows":
        system_str = "win.exe"
    filename = "webapp-builder-" + system_str
    return filename

def get_unzipped_host_filename():
    if platform.system() == "Linux":
        system_str = "linux"
    if platform.system() == "Darwin":
        system_str = "macos"
    if platform.system() == "Windows":
        system_str = "win.exe"
    filename = "webapp-builder-" + system_str + ".v" + __version__
    return filename

def get_unzipped_host_path():
    atri_dir_path = get_atri_dir_path()
    unzipped_file_path = get_unzipped_host_filename()
    return atri_dir_path / unzipped_file_path

def unzip_downloaded_file():
    atri_dir_path = get_atri_dir_path()
    host_filename = get_host_filename()
    path_to_zip_file = atri_dir_path / host_filename
    extract_destination = atri_dir_path
    with zipfile.ZipFile(str(path_to_zip_file), 'r') as zip_ref:
        zip_ref.extractall(str(extract_destination))
    # rename unzipped content
    Path.rename(extract_destination / get_unzipped_remote_filename(), get_unzipped_host_path())
    # change permission to 774 (0o represents octal)
    os.chmod(str(get_unzipped_host_path()), 0o774)

def exe_exists():
    unzipped_exe_path = get_unzipped_host_path()
    if os.path.exists(str(unzipped_exe_path)):
        return True
    return False

def clean_before_download():
    downloaded_zip = get_atri_dir_path() / get_host_filename()
    if os.path.exists(downloaded_zip):
        os.remove(str(downloaded_zip))
    unzipped_path_not_renamed = get_atri_dir_path() / get_unzipped_remote_filename()
    if os.path.exists(str(unzipped_path_not_renamed)):
        os.remove(str(unzipped_path_not_renamed))

def clean_after_download():
    downloaded_zip = get_atri_dir_path() / get_host_filename()
    if os.path.exists(downloaded_zip):
        os.remove(str(downloaded_zip))

def load_exe_if_not_exists():
    create_atri_dir_if_not_exist()
    if not exe_exists():
        clean_before_download()
        print("Downloading atri engine...")
        download_zip()
        print("Extracting...")
        unzip_downloaded_file()
        clean_after_download()
    


import subprocess
from .printd import printd

def pipenv_where():
    child_process = subprocess.Popen(
        ["pipenv", "--where"],
        stdout=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
        stderr=subprocess.PIPE
        )
    child_process.wait()
    if child_process.returncode == 0:
        out = child_process.stdout.read().decode()[:-1]
        printd("Pipenv project root -", out)
        return out
    else:
        print("Failed to detect pipenv project's root directory.")
        printd("[pipenv_where] STDOUT", child_process.stdout.read())
        printd("[pipenv_where] STDERR", child_process.stderr.read())


"""This script is the entrypoint for command line utilities provided in Atri Framework."""
import sys
from atri.errors import NO_CONDA_ENVIRONMENT_FOUND
from atri.utils.is_pkg_installed import is_conda_installed_sync
import click
import asyncio
from .commands.open_editor import run as exe_open_editor
from .commands.connect_local import start_ipc_connection
from .commands.open_exe import open_exe_wrapper
from .commands.load_exe import load_exe_if_not_exists
from .utils.globals import globals
from .commands.check_requisite import check_requisite
if sys.version_info >= (3, 8):
    from asyncio.exceptions import CancelledError
else:
    from asyncio import CancelledError
from .utils.printd import printd
from .find_app_root import find_and_set_app_directory, get_virtualenv_type, is_virtualenv_set, set_virtualenv_type
from . import supported_virt_types
import questionary
from pathlib import Path
from .utils.conda_utils import get_conda_env_list, set_working_env_name, get_active_env_name
from typing import List, Union
from .utils.handle_error import error_to_message
from . import app_config_file
from .commands.create_dockerfile import create_dockerfile_with_pipenv

find_and_set_app_directory()

class VirtTypeQuestion(click.Option):

    def __init__(self, param_decls=None, **attrs):
        click.Option.__init__(self, param_decls, **attrs)
        if not isinstance(self.type, click.Choice):
            raise Exception('ChoiceOption type arg must be click.Choice')

    def prompt_for_value(self, ctx):
        if is_virtualenv_set():
            ctx.obj = {"virt_type": get_virtualenv_type()}
            return None
        val = questionary.select(self.prompt, choices=self.type.choices).unsafe_ask()
        ctx.obj = {"virt_type": val}
        return val

class DefaultEnvNameQuestion(click.Option):

    def __init__(self, param_decls=None, **attrs):
        click.Option.__init__(self, param_decls, **attrs)
        if not isinstance(self.type, click.Choice):
            raise Exception('ChoiceOption type arg must be click.Choice')

    def prompt_for_value(self, ctx):
        if ctx.obj["virt_type"] != "conda":
            return None
        active_env_name = get_active_env_name(str(Path.cwd()))
        if active_env_name != "base":
            return active_env_name
        if len(self.type.choices) == 0:
            print("Please create a conda environment first.")
            exit(NO_CONDA_ENVIRONMENT_FOUND)
        val = questionary.select(self.prompt, choices=self.type.choices).unsafe_ask()
        return val

def get_conda_env_list_if_conda_installed() -> List[str]:
    if is_conda_installed_sync():
        return get_conda_env_list(str(Path.cwd()))
    return []

@click.group()
@click.option("--virt-type", type=click.Choice(supported_virt_types, case_sensitive=False), prompt="Select virtual environment type", cls=VirtTypeQuestion, is_eager=True)
@click.option("--working-env", type=click.Choice(get_conda_env_list_if_conda_installed(), case_sensitive=False), prompt="Select name of the conda virtual env", cls=DefaultEnvNameQuestion)
def main(virt_type: Union[str, None], working_env: Union[str, None]):
    """Open up the visual editor:

        $ atri open editor

    Click publish in the editor once you are done.

    Write your python controllers in controller directory.

    Run dev server during development for automatic refresh whenever you make changes to your code or hit publish in your editor.

        $ atri run dev-server
    
    """
    if virt_type != None:
        set_virtualenv_type(virt_type)
    if working_env != None:
        set_working_env_name(working_env)

@main.group('open')
def open():
    """Open up editor in browser using command:

        $ atri open editor
    """
    pass

@open.command('editor')
@click.option('--e-port', default="4001", help='port on which event server will be attached')
@click.option('--w-port', default="4002", help='port on which file server will be attached to serve static files')
@click.option('--m-port', default="4003", help='port on which manifest server will be attached')
@click.option('--p-port', default="4004", help='port on which publish server will be attached')
@click.option('--d-port', default="4005", help='port on which generate app server will be attached')
@click.option('--u-port', default="4006", help='port on which ipc server will be attached')
@click.option('--c-port', default="4007", help='port on which generated python server will be attached')
@click.option('--no-debug', is_flag = True, default=False, show_default=True, help='run the command in debug mode')
def open_editor(e_port, w_port, m_port, p_port, d_port, u_port, c_port, no_debug):
    """Open up editor in browser using command -

        $ atri open editor --e-port 4001 --w-port 4002 --app-dir atri
    """
    globals["in_debug_mode"] = not no_debug
    app_dir = str(Path.cwd())
    exe_open_editor(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir)

@open.command('exe')
@click.option('--e-port', default="4001", help='port on which event server will be attached')
@click.option('--w-port', default="4002", help='port on which file server will be attached to serve static files')
@click.option('--m-port', default="4003", help='port on which manifest server will be attached')
@click.option('--p-port', default="4004", help='port on which publish server will be attached')
@click.option('--d-port', default="4005", help='port on which generate app server will be attached')
@click.option('--u-port', default="4006", help='port on which ipc server will be attached')
@click.option('--c-port', default="4007", help='port on which generated python server will be attached')
@click.option('--no-debug', is_flag = True, default=False, show_default=True, help='run the command in debug mode')
def open_exe(e_port, w_port, m_port, p_port, d_port, u_port, c_port, no_debug):
    """Open up editor in browser using command -

        $ atri open editor --e-port 4001 --w-port 4002 --app-dir atri
    """
    load_exe_if_not_exists()
    globals["in_debug_mode"] = not no_debug
    app_dir = str(Path.cwd())
    asyncio.run(open_exe_wrapper(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir))

@main.group('connect')
def connect():
    """This command is intended to connect with ipc server when running locally (not inside docker)

        $ atri connect_local
    """
    pass

@connect.command("local")
@click.option('--u-port', default="4006", help='port on which publish server will be attached')
@click.option('--no-debug', is_flag = True, default=False, show_default=True, help='run the command in debug mode')
def connect_local(u_port, no_debug):
    globals["in_debug_mode"] = not no_debug
    app_dir = str(Path.cwd())
    async def check_req_wrapper():
        ok = await check_requisite()
        return ok
    async def connect_local_wrapper():
        sio = await start_ipc_connection(u_port, app_dir)
        await sio.wait()
    async def main_wrapper():
        ok = await check_req_wrapper()
        if ok == 0:
            await connect_local_wrapper()
        else:
            error_to_message(ok)
    asyncio.run(main_wrapper())

@main.command()
@click.option('--e-port', default="4001", help='port on which event server will be attached')
@click.option('--w-port', default="4002", help='port on which file server will be attached to serve static files')
@click.option('--m-port', default="4003", help='port on which manifest server will be attached')
@click.option('--p-port', default="4004", help='port on which publish server will be attached')
@click.option('--d-port', default="4005", help='port on which generate app server will be attached')
@click.option('--u-port', default="4006", help='port on which ipc server will be attached')
@click.option('--c-port', default="4007", help='port on which generated python server will be attached')
@click.option('--debug', is_flag = True, default=False, show_default=True, help='run the command in debug mode')
def start(e_port, w_port, m_port, p_port, d_port, u_port, c_port, debug):
    load_exe_if_not_exists()
    globals["in_debug_mode"] = debug
    app_dir = str(Path.cwd())
    async def check_req_wrapper():
        ok = await check_requisite()
        return ok
    async def connect_local_wrapper():
        sio = await start_ipc_connection(u_port, app_dir)
        await sio.wait()
    async def main_wrapper():
        ok = await check_req_wrapper()
        if ok == 0:
            open_exe_task = asyncio.create_task(
                open_exe_wrapper(e_port, w_port, m_port, p_port, d_port, u_port, c_port, app_dir)
                )
            connect_local_task = asyncio.create_task(
                connect_local_wrapper()
            )
            try:
                await asyncio.wait([open_exe_task, connect_local_task])
            except CancelledError:
                # socket.io AsyncClient throws CancelledError
                # closing stderr to prevent showing error
                sys.stdout.close()
                sys.stderr.close()
            except:
                print("Some error occured while closing atri cli.")
                printd(sys.exc_info())
                exit(1)
            exit(0)
        else:
             error_to_message(ok)
    # Now run the tasks(in the event loop) 
    asyncio.run(main_wrapper())

@main.group()
def check():
    pass

@main.group("create")
def create():
    pass

@create.command("dockerfile")
@click.option('--file', default="Dockerfile", show_default=True, help='Name of the output Dockerfile')
def create_dockerfile(file):
    virt_type = get_virtualenv_type()
    if virt_type == "pipenv":
        create_dockerfile_with_pipenv(file)
    elif virt_type == "conda":
        print("Sorry! We don't support packaging for conda environment yet.")
    elif virt_type == None:
        print("Cannot detect python environment type. Please check {}.".format(app_config_file))
    else:
        print("We don't support {} virtual environment.".format(virt_type))

@check.command()
@click.option('--no-debug', is_flag = True, default=False, show_default=True, help='run the command in debug mode')
def req(no_debug):
    globals["in_debug_mode"] = not no_debug
    ok = asyncio.run(check_requisite())
    if not ok:
        print("If you want to report this issue,",
        "please use our discussion forum https://discuss.atrilabs.com")

if __name__ == '__main__':
    main()


from pathlib import Path
from ..utils.printd import printd
from ..utils.pipenv_utils import pipenv_where
from typing import List
from .. import atri_app_dir, controllers_dir, assets_dir, app_config_file
from ..utils.call_serve import get_common_command

def create_dockerfile_with_pipenv(out_file: str = "Dockerfile"):
    """
    Creates Dockerfile at the root of the project.
    """
    pipenv_root_dir = pipenv_where()
    if pipenv_root_dir == None:
        print("Cannot find pipenv root directory.",
            "Please make sure that you are trying to create Dockerfile",
            "from inside a pipenv project")
        return

    user_dir = "/home/python_user"
    code_dir = user_dir + "/code"
    start_script_path = user_dir + "/start.sh"

    docker_commands: List[str] = []
    docker_commands.append("FROM atrilabs/node16brew")
    docker_commands.append("USER python_user")
    docker_commands.append("RUN mkdir {}".format(code_dir))

    output_path = Path(pipenv_root_dir) / out_file

    if Path.exists(output_path):
        printd("Dockerfile already exists. This command will overwrite the existing file.")

    pipfile_lock_path = Path(pipenv_root_dir) / "Pipfile.lock"
    if Path.exists(pipfile_lock_path):
        docker_commands.append("COPY Pipfile.lock /home/python_user/code")
    
    pipfile_path = Path(pipenv_root_dir) / "Pipfile"
    if Path.exists(pipfile_path):
        docker_commands.append("COPY Pipfile /home/python_user/code")

    # Assuming atri app root is already set as cwd.
    # The atri app root may not be same as pipenv project root.
    assets_path = Path.cwd() / assets_dir
    controllers_path = Path.cwd() / controllers_dir
    atri_app_path = Path.cwd() / atri_app_dir
    app_config_file_path = Path.cwd() / app_config_file

    assets_rel_path = assets_path.relative_to(pipenv_root_dir).as_posix()
    controllers_rel_path = controllers_path.relative_to(pipenv_root_dir).as_posix()
    atri_app_rel_path = atri_app_path.relative_to(pipenv_root_dir).as_posix()
    app_config_file_rel_path = app_config_file_path.relative_to(pipenv_root_dir).as_posix()

    docker_commands.append("COPY {} {}/{}".format(assets_rel_path, code_dir, assets_rel_path))
    docker_commands.append("COPY {} {}/{}".format(controllers_rel_path, code_dir, controllers_rel_path))
    docker_commands.append("COPY {} {}/{}".format(atri_app_rel_path, code_dir, atri_app_rel_path))
    docker_commands.append("COPY {} {}/{}".format(app_config_file_rel_path, code_dir, app_config_file_rel_path))

    # TODO: add command to execute python controller server & node server
    start_script_commands: List[str] = []
    start_script_commands.append("cd {}".format(code_dir))
    start_script_commands.append("pipenv run {} &".format(get_common_command()))
    start_script_commands.append("cd {}/{}".format(code_dir, atri_app_rel_path))
    start_script_commands.append("yarn run server".format(get_common_command()))  

    docker_commands.append('RUN echo "{}" > {}'.format("\\n".join(start_script_commands), start_script_path))

    # change ownership of /home/python_user
    docker_commands.append("USER root")
    docker_commands.append("RUN chown -R python_user {}".format(user_dir))

    # install npm packages
    docker_commands.append("WORKDIR {}/{}".format(code_dir, atri_app_rel_path))
    docker_commands.append("RUN yarn install && yarn run build && yarn run buildServer")

    # install python packages
    docker_commands.append("RUN python3 -m pip install pipenv")
    docker_commands.append("WORKDIR {}".format(code_dir))
    docker_commands.extend([
        "ENV LANG C.UTF-8",
        "ENV LC_ALL C.UTF-8",
        "ENV PYTHONDONTWRITEBYTECODE 1",
        "ENV PYTHONFAULTHANDLER 1"
        ])
    docker_commands.append("RUN PIPENV_VENV_IN_PROJECT=1 pipenv install")

    docker_commands.append("USER python_user")
    docker_commands.append('CMD ["{}"]'.format(start_script_path))

    with open(output_path, "w") as f:
        f.write("\n".join(docker_commands))
